class ThrottleBackendBase:
    def incr_bucket(self, zone_name, bucket_key, bucket_num, bucket_num_next, bucket_span, cost=1):
        """
        Increments the limit for the given bucket.

        @returns: the new value of the bucket, post-increment
        """
        raise NotImplementedError



import time
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings

from throttle.zones.remoteip import RemoteIP
from throttle.exceptions import ThrottleZoneNotDefined, ThrottleImproperlyConfigured, RateLimitExceeded
from throttle.utils import load_class_from_path, serialize_bucket_key
from throttle.backends import get_backend

THROTTLE_ENABLED = getattr(settings, 'THROTTLE_ENABLED', not settings.DEBUG)

_THROTTLE_ZONES = {}


class ThrottleZone(object):
    def __init__(self, zone_name, vary_with, **config):
        self._zone_name = zone_name
        self.vary = vary_with(**config)
        self.config = config
        self.get_timestamp = lambda: int(time.time())

        try:
            self.bucket_interval = int(config['BUCKET_INTERVAL'])
            if self.bucket_interval <= 0:
                raise ValueError
        except KeyError:
            raise ThrottleImproperlyConfigured('THROTTLE_ZONE[\'%s\'] missing BUCKET_INTERVAL parameter)' % zone_name)
        except ValueError:
            raise ThrottleImproperlyConfigured('THROTTLE_ZONE[\'%s\'][\'BUCKET_INTERVAL\'] must be > 0' % zone_name)

        try:
            self.num_buckets = int(config['NUM_BUCKETS'])
            if self.num_buckets < 2:
                raise ValueError
        except KeyError:
            raise ThrottleImproperlyConfigured('THROTTLE_ZONE[\'%s\'] missing NUM_BUCKETS parameter)' % zone_name)
        except ValueError:
            raise ThrottleImproperlyConfigured('THROTTLE_ZONE[\'%s\'][\'NUM_BUCKETS\'] must be >= 2' % zone_name)

        try:
            self.bucket_capacity = int(config['BUCKET_CAPACITY'])
        except KeyError:
            raise ThrottleImproperlyConfigured('THROTTLE_ZONE[\'%s\'] missing BUCKET_CAPACITY parameter)' % zone_name)
        except ValueError:
            raise ThrottleImproperlyConfigured('THROTTLE_ZONE[\'%s\'][\'BUCKET_CAPACITY\'] must be an int' % zone_name)

        self.bucket_span = self.bucket_interval * self.num_buckets

    def process_view(self, request, view_func, view_args, view_kwargs):
        # If THROTTLE_ENABLED is False, just return the response from the view.
        if not THROTTLE_ENABLED:
            return view_func(request, *view_args, **view_kwargs)

        # if func is a class view the request is the instance of the class
        view_class = request
        request = getattr(request, 'request', request)

        bucket_key = serialize_bucket_key(
            self.vary.get_bucket_key(request, view_func, view_args, view_kwargs)
        )

        # Calculate the bucket offset to increment
        timestamp = self.get_timestamp()
        bucket_num = int((timestamp % self.bucket_span) / self.bucket_interval)
        bucket_num_next = (bucket_num + 1) % self.num_buckets

        # Tell the backing store to increment the count
        new_value = get_backend().incr_bucket(self.name, bucket_key, bucket_num, bucket_num_next, self.bucket_span)

        # Has the bucket capacity been exceeded?
        if new_value > self.bucket_capacity:
            raise RateLimitExceeded(self.name)

        num_remaining = self.bucket_capacity - new_value

        # Execute the Django view. Add a few attributes to the response object.
        response = view_func(view_class, *view_args, **view_kwargs)
        response.throttle_limit = self.bucket_capacity
        response.throttle_remaining = num_remaining

        # Perform additional processing on the response object
        # TODO: Make this better
        response = self.process_response(request, response, remaining=num_remaining)

        return response

    def process_response(self, request, response, remaining=0):
        # TODO: Make this configurable
        response['X-Request-Limit-Limit'] = self.bucket_capacity
        response['X-Request-Limit-Remaining'] = remaining
        return response

    @property
    def name(self):
        return self._zone_name


def _load_zone(zone_name, **config):
    vary_klass = load_class_from_path(config['VARY'])
    return ThrottleZone(zone_name, vary_klass, **config)


def get_zone(zone_name):
    try:
        return _THROTTLE_ZONES[zone_name]
    except KeyError:
        try:
            throttle_zone = settings.THROTTLE_ZONES[zone_name]
            zone = _load_zone(zone_name, **throttle_zone)
            _THROTTLE_ZONES[zone_name] = zone
            return zone
        except AttributeError:
            raise ImproperlyConfigured('@throttle is used, but settings.THROTTLE_ZONES is undefined')
        except KeyError:
            raise ThrottleZoneNotDefined(zone_name)



# -*- coding: utf-8 -*-
import hashlib

from throttle.backends.base import ThrottleBackendBase
try:
    import redis
    from redis.exceptions import NoScriptError
except ImportError:
    from throttle.exceptions import ThrottleImproperlyConfigured
    raise ThrottleImproperlyConfigured("django-throttle-requests is configured to use redis, but redis-py is not installed!")

# Lua script to update bucket data atomically.
# In general, lua scripts should be used instead of Redis transactions to ensure atomicity. Transactions may be
# deprecated at some point. Also, nutcracker does not support transactions but does support scripting
# as long as all keys used by the script hash to the same backend. The same limitation applies to Redis Cluster.
#
# Script takes 1 key and 4 arguments: <bucket_num>, <bucket_num_next>, <bucket_span>, <cost>
INCR_BUCKET_SCRIPT = """
local newval = redis.call('hincrby', KEYS[1], ARGV[1], ARGV[4])
redis.call('hdel', KEYS[1], ARGV[2])
redis.call('expire', KEYS[1], ARGV[3])
return newval
"""

INCR_BUCKET_SCRIPT_SHA1 = hashlib.sha1(INCR_BUCKET_SCRIPT.encode('utf-8')).hexdigest()

from django.conf import settings

class RedisBackend(ThrottleBackendBase):
    def __init__(self):
        
        THROTTLE_REDIS_HOST = getattr(settings, 'THROTTLE_REDIS_HOST', 'localhost')
        THROTTLE_REDIS_PORT = getattr(settings, 'THROTTLE_REDIS_PORT', 6379)
        THROTTLE_REDIS_DB = getattr(settings, 'THROTTLE_REDIS_DB', 0)
        THROTTLE_REDIS_AUTH = getattr(settings, 'THROTTLE_REDIS_AUTH', None)

        self.pool = redis.ConnectionPool(host=THROTTLE_REDIS_HOST, port=THROTTLE_REDIS_PORT, db=THROTTLE_REDIS_DB, password=THROTTLE_REDIS_AUTH)  #TODO: Parameterize connection parameters

    def incr_bucket(self, zone_name, bucket_key, bucket_num, bucket_num_next, bucket_span, cost=1):
        conn = redis.Redis(connection_pool=self.pool)

        bucket_cache_key = "%s:%s" % (zone_name, bucket_key)

        # Don't want to use redispy's `register_script` command here, because it uses SCRIPT LOAD, which isn't compatible
        # with nutcracker or Redis Cluster.
        try:
            try:
                return conn.evalsha(INCR_BUCKET_SCRIPT_SHA1, 1, bucket_cache_key, bucket_num, bucket_num_next, bucket_span, cost)
            except NoScriptError:
                return conn.eval(INCR_BUCKET_SCRIPT, 1, bucket_cache_key, bucket_num, bucket_num_next, bucket_span, cost)

        except redis.ConnectionError:
            return cost



import functools

from throttle.zones import get_zone


def throttle(view_func=None, zone='default'):
    def decorator(func):

        @functools.wraps(func, assigned=functools.WRAPPER_ASSIGNMENTS)
        def _wrapped_view(request, *args, **kwargs):
            # Get zone from cache
            _throttle_zone = get_zone(zone)

            # raises an exception if the rate limit is exceeded
            response = _throttle_zone.process_view(request, func, args, kwargs)
            return response

        # Validate the rate limiter bucket
        _zone = get_zone(zone)

        if func:
            setattr(_wrapped_view, 'throttle_zone', _zone)
            return _wrapped_view
        return _wrapped_view

    # Validate the rate limiter bucket
    _zone = get_zone(zone)
    if view_func:
        setattr(view_func, 'throttle_zone', _zone)
        return decorator(view_func)
    return decorator



class RemoteIP:
    def __init__(self, **config):
        pass

    def get_bucket_key(self, request, view_func, view_args, view_kwargs):
        """
        Return our best crack at the remote IP address
        """
        # Handle Load Balancer case
        # Load balancer returns usually in format of 'xx.xx.xx.xx, yy.yy.yy.yy, zz.zz.zz.zz'
        # where xx.xx.xx.xx is the actual ip while others vary on the request
        ip_string = request.META.get('HTTP_X_FORWARDED_FOR', "")
        ip_string = ip_string.split(',')[0]
        return ip_string or request.META.get('REMOTE_ADDR')



from django.core.exceptions import ImproperlyConfigured, PermissionDenied

class RateLimitExceeded(PermissionDenied):
    pass

class ThrottleZoneNotDefined(ImproperlyConfigured):
    pass

class ThrottleImproperlyConfigured(ImproperlyConfigured):
    pass


from django.core.cache import cache
from throttle.backends.base import ThrottleBackendBase


class CacheBackend(ThrottleBackendBase):
    def __init__(self):
        self.cache = cache

    def incr_bucket(self, zone_name, bucket_key, bucket_num, bucket_num_next, bucket_span, cost=1):
        bucket_cache_key = "%s:%s:%s" % (zone_name, bucket_key, bucket_num)
        next_bucket_cache_key = "%s:%s:%s" % (zone_name, bucket_key, bucket_num_next)

        # Increment the value of the current bucket
        try:
            new_val = self.cache.incr(bucket_cache_key, cost)
        except ValueError:
            # bucket_cache_key not in the cache
            self.cache.add(bucket_cache_key, cost, bucket_span)
            new_val = cost

        # Reset the value of the next bucket
        self.cache.set(next_bucket_cache_key, 0, bucket_span)

        return new_val



from django.core.exceptions import ImproperlyConfigured
from django.conf import settings

from throttle.utils import load_class_from_path

THROTTLE_BACKEND = getattr(settings, 'THROTTLE_BACKEND', {})
_backend = None

def load_backend_from_path(classpath):
    klass = load_class_from_path(classpath)
    try:
        return klass()
    except TypeError:
        raise ImproperlyConfigured("%s is not callable" % (klass.__name__))

def get_backend():
    global _backend
    if _backend:
        return _backend

    if THROTTLE_BACKEND:
        _backend = load_backend_from_path(settings.THROTTLE_BACKEND)
        return _backend
    else:
        raise ImproperlyConfigured('@throttle was used, but settings.THROTTLE_BACKEND is not set')


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup
try:
    import multiprocessing # Suppresses a confusing but harmless warning when running ./setup.py test
    import logging
except ImportError:
    # multiprocessing introduced in Python 2.6
    pass

from throttle import __version__

long_description = open('README.rst').read()

setup(
    name="django-throttle-requests",
    description="A Django framework for application-layer rate limiting",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    packages=find_packages(),
    url="https://github.com/sobotklp/django-throttle-requests",
    version=__version__,
    author='Lewis Sobotkiewicz',
    author_email='lewis.sobot@gmail.com',
    install_requires=[
        'Django>=1.7',
    ],
    license='MIT',
    test_suite='runtests.runtests',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from hashlib import sha256

from django.core.exceptions import ImproperlyConfigured
try:
    from importlib import import_module
except ImportError:
    from django.utils.importlib import import_module

def load_class_from_path(class_path):
    # Split the class into a <module, classname> pair
    try:
        modulename, classname = class_path.rsplit('.', 1)
    except ValueError:
        raise ImproperlyConfigured("'%s' isn't a valid module name" % (class_path))

    # Attempt to load the module
    try:
        module = import_module(modulename)
    except (ImportError,):
        raise ImproperlyConfigured("Error importing module '%s'" % (modulename))

    # Attempt to reference the class
    try:
        return getattr(module, classname)
    except AttributeError:
        raise ImproperlyConfigured("Module %s has no class '%s'" % (modulename, classname))

# Memcached can't take key with the spaces. This would be happen when
#   you use more than one reverse proxy.
def serialize_bucket_key(bucket_key):
    return sha256(bucket_key.encode('utf-8')).hexdigest()
